@@ -6,7 +6,7 @@
  * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
- *	$Header: /cvsroot/pgsql-server/src/backend/parser/analyze.c,v 1.290.2.1 2003/11/05 22:00:52 tgl Exp $
+ *	$Header: /cvsroot/pgsql/src/backend/parser/analyze.c,v 1.290.2.4 2006/10/11 20:03:26 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -440,6 +440,18 @@ transformStmt(ParseState *pstate, Node *parseTree,
 	result->querySource = QSRC_ORIGINAL;
 	result->canSetTag = true;
 
+	/*
+	 * Check that we did not produce too many resnos; at the very
+	 * least we cannot allow more than 2^16, since that would exceed
+	 * the range of a AttrNumber. It seems safest to use
+	 * MaxTupleAttributeNumber.
+	 */
+	if (pstate->p_next_resno - 1 > MaxTupleAttributeNumber)
+		ereport(ERROR,
+				(errcode(ERRCODE_PROGRAM_LIMIT_EXCEEDED),
+				 errmsg("target lists can have at most %d entries",
+						MaxTupleAttributeNumber)));
+
 	return result;
 }
 
@@ -486,6 +498,8 @@ transformInsertStmt(ParseState *pstate, InsertStmt *stmt,
 					List **extras_before, List **extras_after)
 {
 	Query	   *qry = makeNode(Query);
+	Query	   *selectQuery = NULL;
+	bool		copy_up_hack = false;
 	List	   *sub_rtable;
 	List	   *sub_namespace;
 	List	   *icolumns;
@@ -540,7 +554,6 @@ transformInsertStmt(ParseState *pstate, InsertStmt *stmt,
 		 * be able to see.
 		 */
 		ParseState *sub_pstate = make_parsestate(pstate);
-		Query	   *selectQuery;
 		RangeTblEntry *rte;
 		RangeTblRef *rtr;
 
@@ -587,7 +600,7 @@ transformInsertStmt(ParseState *pstate, InsertStmt *stmt,
 		Assert(rte == rt_fetch(rtr->rtindex, pstate->p_rtable));
 		pstate->p_joinlist = lappend(pstate->p_joinlist, rtr);
 
-		/*
+		/*----------
 		 * Generate a targetlist for the INSERT that selects all the
 		 * non-resjunk columns from the subquery.  (We need this to be
 		 * separate from the subquery's tlist because we may add columns,
@@ -597,8 +610,9 @@ transformInsertStmt(ParseState *pstate, InsertStmt *stmt,
 		 * are copied up as-is rather than being referenced as subquery
 		 * outputs.  This is to ensure that when we try to coerce them to
 		 * the target column's datatype, the right things happen (see
-		 * special cases in coerce_type).  Otherwise, this fails: INSERT
-		 * INTO foo SELECT 'bar', ... FROM baz
+		 * special cases in coerce_type).  Otherwise, this fails:
+		 *		INSERT INTO foo SELECT 'bar', ... FROM baz
+		 *----------
 		 */
 		qry->targetList = NIL;
 		foreach(tl, selectQuery->targetList)
@@ -610,9 +624,12 @@ transformInsertStmt(ParseState *pstate, InsertStmt *stmt,
 			if (resnode->resjunk)
 				continue;
 			if (tle->expr &&
-				(IsA(tle->expr, Const) ||IsA(tle->expr, Param)) &&
+				(IsA(tle->expr, Const) || IsA(tle->expr, Param)) &&
 				exprType((Node *) tle->expr) == UNKNOWNOID)
+			{
 				expr = tle->expr;
+				copy_up_hack = true;
+			}
 			else
 				expr = (Expr *) makeVar(rtr->rtindex,
 										resnode->resno,
@@ -686,6 +703,28 @@ transformInsertStmt(ParseState *pstate, InsertStmt *stmt,
 				(errcode(ERRCODE_SYNTAX_ERROR),
 			 errmsg("INSERT has more target columns than expressions")));
 
+	/*
+	 * If we copied up any unknown Params (see HACK above) then their
+	 * resolved types need to be propagated into the Resdom nodes of
+	 * the sub-INSERT's tlist.  One hack begets another :-(
+	 */
+	if (copy_up_hack)
+	{
+		foreach(tl, selectQuery->targetList)
+		{
+			TargetEntry *tle = (TargetEntry *) lfirst(tl);
+			Resdom	   *resnode = tle->resdom;
+
+			if (resnode->resjunk)
+				continue;
+			if (resnode->restype == UNKNOWNOID)
+			{
+				resnode->restype = exprType((Node *) tle->expr);
+				resnode->restypmod = exprTypmod((Node *) tle->expr);
+			}
+		}
+	}
+
 	/* done building the range table and jointree */
 	qry->rtable = pstate->p_rtable;
 	qry->jointree = makeFromExpr(pstate->p_joinlist, NULL);
@@ -1201,7 +1240,7 @@ transformInhRelation(ParseState *pstate, CreateStmtContext *cxt,
 	constr = tupleDesc->constr;
 
 	/*
-	 * Insert the inherited attributes into the cxt for the new table
+	 * Insert the copied attributes into the cxt for the new table
 	 * definition.
 	 */
 	for (parent_attno = 1; parent_attno <= tupleDesc->natts;
@@ -1219,7 +1258,7 @@ transformInhRelation(ParseState *pstate, CreateStmtContext *cxt,
 			continue;
 
 		/*
-		 * Create a new inherited column.
+		 * Create a new column, which is marked as NOT inherited.
 		 *
 		 * For constraints, ONLY the NOT NULL constraint is inherited by the
 		 * new column definition per SQL99.
@@ -1231,7 +1270,7 @@ transformInhRelation(ParseState *pstate, CreateStmtContext *cxt,
 		typename->typmod = attribute->atttypmod;
 		def->typename = typename;
 		def->inhcount = 0;
-		def->is_local = false;
+		def->is_local = true;
 		def->is_not_null = attribute->attnotnull;
 		def->raw_default = NULL;
 		def->cooked_default = NULL;
